/**
 * Logback: the reliable, generic, fast and flexible logging framework.
 * Copyright (C) 1999-2015, QOS.ch. All rights reserved.
 *
 * This program and the accompanying materials are dual-licensed under
 * either the terms of the Eclipse Public License v1.0 as published by
 * the Eclipse Foundation
 *
 *   or (per the licensee's choosing)
 *
 * under the terms of the GNU Lesser General Public License version 2.1
 * as published by the Free Software Foundation.
 */
package org.apache.juli.logging.ch.qos.logback.core.joran;

import java.util.HashMap;
import java.util.Map;

import org.apache.juli.logging.ch.qos.logback.core.Appender;
import org.apache.juli.logging.ch.qos.logback.core.Context;
import org.apache.juli.logging.ch.qos.logback.core.joran.action.AppenderAction;
import org.apache.juli.logging.ch.qos.logback.core.joran.action.AppenderRefAction;
import org.apache.juli.logging.ch.qos.logback.core.joran.action.ContextPropertyAction;
import org.apache.juli.logging.ch.qos.logback.core.joran.action.ConversionRuleAction;
import org.apache.juli.logging.ch.qos.logback.core.joran.action.DefinePropertyAction;
import org.apache.juli.logging.ch.qos.logback.core.joran.action.EventEvaluatorAction;
import org.apache.juli.logging.ch.qos.logback.core.joran.action.ImplicitModelAction;
import org.apache.juli.logging.ch.qos.logback.core.joran.action.NewRuleAction;
import org.apache.juli.logging.ch.qos.logback.core.joran.action.ParamAction;
import org.apache.juli.logging.ch.qos.logback.core.joran.action.PropertyAction;
import org.apache.juli.logging.ch.qos.logback.core.joran.action.ShutdownHookAction;
import org.apache.juli.logging.ch.qos.logback.core.joran.action.StatusListenerAction;
import org.apache.juli.logging.ch.qos.logback.core.joran.action.TimestampAction;
import org.apache.juli.logging.ch.qos.logback.core.joran.spi.ElementSelector;
import org.apache.juli.logging.ch.qos.logback.core.joran.spi.InterpretationContext;
import org.apache.juli.logging.ch.qos.logback.core.joran.spi.RuleStore;
import org.apache.juli.logging.ch.qos.logback.core.joran.spi.SaxEventInterpreter;
import org.apache.juli.logging.ch.qos.logback.core.model.DefineModel;
import org.apache.juli.logging.ch.qos.logback.core.model.EventEvaluatorModel;
import org.apache.juli.logging.ch.qos.logback.core.model.ImplicitModel;
import org.apache.juli.logging.ch.qos.logback.core.model.ParamModel;
import org.apache.juli.logging.ch.qos.logback.core.model.PropertyModel;
import org.apache.juli.logging.ch.qos.logback.core.model.ShutdownHookModel;
import org.apache.juli.logging.ch.qos.logback.core.model.StatusListenerModel;
import org.apache.juli.logging.ch.qos.logback.core.model.TimestampModel;
import org.apache.juli.logging.ch.qos.logback.core.model.processor.DefaultProcessor;
import org.apache.juli.logging.ch.qos.logback.core.model.processor.DefineModelHandler;
import org.apache.juli.logging.ch.qos.logback.core.model.processor.EventEvaluatorModelHandler;
import org.apache.juli.logging.ch.qos.logback.core.model.processor.ImplicitModelHandler;
import org.apache.juli.logging.ch.qos.logback.core.model.processor.PropertyModelHandler;
import org.apache.juli.logging.ch.qos.logback.core.model.processor.ShutdownHookModelHandler;
import org.apache.juli.logging.ch.qos.logback.core.model.processor.StatusListenerModelHandler;
import org.apache.juli.logging.ch.qos.logback.core.model.processor.TimestampModelHandler;
import org.apache.juli.logging.ch.qos.logback.core.spi.AppenderAttachable;


// Based on 310985 revision 310985 as attested by http://tinyurl.com/8njps
// see also http://tinyurl.com/c2rp5

/**
 * A JoranConfiguratorBase lays most of the groundwork for concrete
 * configurators derived from it. Concrete configurators only need to implement
 * the {@link #addInstanceRules} method.
 * <p>
 * A JoranConfiguratorBase instance should not be used more than once to
 * configure a Context.
 *
 * @author Ceki G&uuml;lc&uuml;
 */
abstract public class JoranConfiguratorBase<E> extends GenericConfigurator {

    @Override
    protected void addInstanceRules(RuleStore rs) {

        // is "configuration/variable" referenced in the docs?
        rs.addRule(new ElementSelector("configuration/variable"), new PropertyAction());
        rs.addRule(new ElementSelector("configuration/property"), new PropertyAction());

        rs.addRule(new ElementSelector("configuration/substitutionProperty"), new PropertyAction());

        rs.addRule(new ElementSelector("configuration/timestamp"), new TimestampAction());
        rs.addRule(new ElementSelector("configuration/shutdownHook"), new ShutdownHookAction());
        rs.addRule(new ElementSelector("configuration/define"), new DefinePropertyAction());
        rs.addRule(new ElementSelector("configuration/evaluator"), new EventEvaluatorAction());

        // the contextProperty pattern is deprecated. It is undocumented
        // and will be dropped in future versions of logback
        rs.addRule(new ElementSelector("configuration/contextProperty"), new ContextPropertyAction());

        rs.addRule(new ElementSelector("configuration/conversionRule"), new ConversionRuleAction());

        rs.addRule(new ElementSelector("configuration/statusListener"), new StatusListenerAction());

        rs.addRule(new ElementSelector("configuration/appender"), new AppenderAction());
        rs.addRule(new ElementSelector("configuration/appender/appender-ref"), new AppenderRefAction());
        rs.addRule(new ElementSelector("configuration/newRule"), new NewRuleAction());
        
        rs.addRule(new ElementSelector("*/param"), new ParamAction());
        
    }

    @Override
    protected void addImplicitRules(SaxEventInterpreter interpreter) {
        // The following line adds the capability to parse nested components
//        NestedComplexPropertyIA nestedComplexPropertyIA = new NestedComplexPropertyIA(getBeanDescriptionCache());
//        nestedComplexPropertyIA.setContext(context);
//        interpreter.addImplicitAction(nestedComplexPropertyIA);
//
//        NestedBasicPropertyIA nestedBasicIA = new NestedBasicPropertyIA(getBeanDescriptionCache());
//        nestedBasicIA.setContext(context);
//        interpreter.addImplicitAction(nestedBasicIA);
        
        ImplicitModelAction implicitRuleModelAction = new  ImplicitModelAction();
        implicitRuleModelAction.setContext(context);
        interpreter.addImplicitAction(implicitRuleModelAction);
    }

    @Override
    protected void buildInterpreter() {
        super.buildInterpreter();
        Map<String, Object> omap = interpreter.getInterpretationContext().getObjectMap();
        omap.put(JoranConstants.APPENDER_BAG, new HashMap<String, Appender<?>>());
        omap.put(JoranConstants.APPENDER_REF_BAG, new HashMap<String, AppenderAttachable<?>>());
        
        //omap.put(ActionConst.FILTER_CHAIN_BAG, new HashMap());
    }

    public InterpretationContext getInterpretationContext() {
        return interpreter.getInterpretationContext();
    }

    @Override
    protected DefaultProcessor buildDefaultProcessor(Context context, InterpretationContext interpretationContext) {
        DefaultProcessor defaultProcessor = super.buildDefaultProcessor(context, interpretationContext);
        defaultProcessor.addHandler(ShutdownHookModel.class, ShutdownHookModelHandler.class);
        defaultProcessor.addHandler(EventEvaluatorModel.class, EventEvaluatorModelHandler.class);
        defaultProcessor.addHandler(DefineModel.class, DefineModelHandler.class);
        
        defaultProcessor.addHandler(ParamModel.class, ParamModelHandler.class);
        defaultProcessor.addHandler(PropertyModel.class, PropertyModelHandler.class);
        defaultProcessor.addHandler(TimestampModel.class, TimestampModelHandler.class);
        defaultProcessor.addHandler(StatusListenerModel.class, StatusListenerModelHandler.class);
        defaultProcessor.addHandler(ImplicitModel.class, ImplicitModelHandler.class);

        return defaultProcessor;
    }

}
